import pandas as pd
import numpy as np
import geopandas as gpd
from shapely.geometry import MultiPoint
from itertools import product
import glob
from datetime import timedelta, datetime
from shapely.wkt import loads
from copy import deepcopy
from scipy.spatial.distance import directed_hausdorff

import config as cfg

def main():

    # step 1 combine cases
    concepts = list(cfg.experiment_cases['concepts'].keys())

    exp_cases = list(
        product(
            cfg.experiment_cases['concepts'].keys(),
            cfg.experiment_cases['densities'],
            cfg.experiment_cases['clusters'],
            cfg.experiment_cases['replanlimits'],
            cfg.experiment_cases['replanratios'],
            cfg.experiment_cases['graphweights'],
            cfg.experiment_cases['seeds'],
            cfg.experiment_cases['CR'],
        )
    )


    # initialize dataframe to save general data
    columns = ['concept', 'density', 'cluster', 'replanlimit', 'replanratio', 'graphweights', 'seed', 'CR'] \
            + list(cfg.safety_dict['CONFLOG']['metrics'].keys())    \
            + list(cfg.safety_dict['LOSLOG']['metrics'].keys())    \
            + list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())    \
            + list(cfg.flow_dict['FLOWLOG']['metrics'].keys())    \
            + list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys()) \
            + list(cfg.replan_dict['REPLANLOG']['metrics'].keys())


    data = {column: [] for column in columns}

    # LOG processing
    for concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase in exp_cases:   


        # start with replanlog
        replanresult = replanlog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, 'REPLANLOG')
        ac_number_of_replanning_aircraft, ac_number_of_not_replanning_aircraft, ac_number_of_attempted_replans, \
        executed_replans_per_flight, attempted_replans_per_flight, number_of_weird_situations, mission_acids = replanresult     

        #  get number of conflicts and intrusions
        n_confs_total = conflog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, 'CONFLOG', mission_acids)
        n_intrusions_total = loslog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, 'LOSLOG', mission_acids)

        # Next step is to get some data about the FLSTLOG
        n_flights, flight_time, flight_distance = flstlog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, 'FLSTLOG', mission_acids)
                
        # Now define the flowlog
        flowlogresult = flowlog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, 'FLOWLOG')
        mean_attempted_replans, mean_succesful_replans, percent_of_attempted, percent_of_total, n_succesful_replans = flowlogresult 
        

        # # now do cluster analysis
        clusterresult = clusterlog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, 'CLUSTERLOG')
        percent_medium_clusters, percent_high_clusters, percent_medium_airspace, \
        percent_high_airspace, percent_clustered_airspace, \
        hausdorff_distances_medium, hausdorff_distances_high, hausdorff_distances_all, \
        overlap_percentages_medium, overlap_percentages_high, overlap_percentages_all = clusterresult


        # do a reglog analysis
        reglog_gdf = reglog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, 'REGLOG', mission_acids)


        # extend data
        data['concept'].append(concept)
        data['density'].append(density)
        data['cluster'].append(cluster)
        data['replanlimit'].append(replanlimit)
        data['replanratio'].append(replanratio)
        data['graphweights'].append(graphweights)
        data['seed'].append(seed)
        data['CR'].append(CRcase)

        data[list(cfg.safety_dict['CONFLOG']['metrics'].keys())[0]].append(n_confs_total)
        data[list(cfg.safety_dict['CONFLOG']['metrics'].keys())[1]].append(n_confs_total / n_flights)
        data[list(cfg.safety_dict['CONFLOG']['metrics'].keys())[2]].append(n_confs_total/flight_distance)
        data[list(cfg.safety_dict['CONFLOG']['metrics'].keys())[3]].append(0.0)
        data[list(cfg.safety_dict['CONFLOG']['metrics'].keys())[4]].append(0.0)

        data[list(cfg.safety_dict['LOSLOG']['metrics'].keys())[0]].append(n_intrusions_total)
        data[list(cfg.safety_dict['LOSLOG']['metrics'].keys())[1]].append(n_intrusions_total / n_flights)
        data[list(cfg.safety_dict['LOSLOG']['metrics'].keys())[2]].append(n_intrusions_total/flight_distance)
        data[list(cfg.safety_dict['LOSLOG']['metrics'].keys())[3]].append(0.0)
        data[list(cfg.safety_dict['LOSLOG']['metrics'].keys())[4]].append(0.0)


        data[list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())[0]].append(n_flights)
        data[list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())[1]].append(flight_time)
        data[list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())[2]].append(flight_distance)
        data[list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())[3]].append(flight_time / n_flights)
        data[list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())[4]].append(flight_distance / n_flights)
        data[list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())[5]].append(0.0)
        data[list(cfg.efficency_dict['FLSTLOG']['metrics'].keys())[6]].append(0.0)

        data[list(cfg.flow_dict['FLOWLOG']['metrics'].keys())[0]].append(mean_attempted_replans)
        data[list(cfg.flow_dict['FLOWLOG']['metrics'].keys())[1]].append(mean_succesful_replans)
        data[list(cfg.flow_dict['FLOWLOG']['metrics'].keys())[2]].append(percent_of_attempted)
        data[list(cfg.flow_dict['FLOWLOG']['metrics'].keys())[3]].append(percent_of_total)
        data[list(cfg.flow_dict['FLOWLOG']['metrics'].keys())[4]].append(n_succesful_replans/n_flights)

        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[0]].append(percent_medium_clusters)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[1]].append(percent_high_clusters)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[2]].append(percent_medium_airspace)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[3]].append(percent_high_airspace)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[4]].append(percent_clustered_airspace)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[5]].append(hausdorff_distances_medium)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[6]].append(hausdorff_distances_high)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[7]].append(hausdorff_distances_all)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[8]].append(overlap_percentages_medium)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[9]].append(overlap_percentages_high)
        data[list(cfg.cluster_dict['CLUSTERLOG']['metrics'].keys())[10]].append(overlap_percentages_all)

        data[list(cfg.replan_dict['REPLANLOG']['metrics'].keys())[0]].append(executed_replans_per_flight)


    # create the general safety dataframe
    df = pd.DataFrame(data=data)

    # after everything has been added get some percentages
    for concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase in exp_cases:        

        if concept != 'baseline':
            continue

        # case dataframe
        case_df = df[(df['density'] == density) & (df['cluster'] == cluster) & (df['seed'] == seed)]
        
        if len(case_df) == 0:
            continue
        
        # print(concept, density, cluster, seed, CRcase)
        # calculate flight time percentages as a function of baseline
        flight_time = case_df[case_df['concept'] == 'baseline']['Flight time per flight'].values[0]
        flight_distance = case_df[case_df['concept'] == 'baseline']['Distance per flight'].values[0] 

        time_df = case_df['Flight time per flight'] / flight_time
        distance_df = case_df['Distance per flight'] / flight_distance

        df.loc[time_df.index,'Flight time percentage'] = time_df*100
        df.loc[time_df.index,'Distance 2D percentage'] = distance_df*100

        # calculte intrusion percentage
        intrusions_per_flight = case_df[case_df['concept'] == 'baseline']['Intrusions per flight'].values[0]
        intrusions_per_distance = case_df[case_df['concept'] == 'baseline']['Intrusions per distance'].values[0]

        intrusion_df_flight = case_df['Intrusions per flight'] / intrusions_per_flight
        intrusion_df_distance = case_df['Intrusions per distance'] / intrusions_per_distance

        df.loc[time_df.index,'Intrusions per flight percentage'] = intrusion_df_flight*100
        df.loc[time_df.index,'Intrusions per distance percentage'] = intrusion_df_distance*100

        # calculte conflict percentage
        conflicts_per_flight = case_df[case_df['concept'] == 'baseline']['Conflicts per flight'].values[0]
        conflicts_per_distance = case_df[case_df['concept'] == 'baseline']['Conflicts per distance'].values[0]

        conflict_df_flight = case_df['Conflicts per flight'] / conflicts_per_flight
        conflict_df_distance = case_df['Conflicts per distance'] / conflicts_per_distance

        df.loc[time_df.index,'Conflicts per flight percentage'] = conflict_df_flight*100
        df.loc[time_df.index,'Conflicts per distance percentage'] = conflict_df_distance*100

        

    # save the dataframes
    df.to_csv(f'dataframes/{cfg.dataframe_name}.csv', index=False)

def scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype):

    if concept == 'baseline':
        scenario = f'{concept}_traf{density}_clust4000_replanlimit30_replanratio0.5_graphweights1-1.5-1.5_seed{seed}_CR{CRcase}'
        scenario_glob = cfg.baseline_log_location + f'/{logtype}_' + scenario + '*.log'

    else:
        scenario = f'{concept}_traf{density}_clust{cluster}_replanlimit{replanlimit}_replanratio{replanratio}_graphweights{graphweights}_seed{seed}_CR{CRcase}'
        scenario_glob = cfg.log_location + f'/{logtype}_' + scenario + '*.log'

    scenario_path = glob.glob(scenario_glob)[0]
    
    return scenario_path

def conflog(concept:str, density:str, cluster:str, replanlimit:str, replanratio:str, graphweights:str, seed:str, CRcase:str, logtype:str, mission_acids) -> int:


    scenario_path = scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype)

    df = pd.read_csv(
        scenario_path,
        skiprows=9,
        names=cfg.safety_dict[logtype]['columns']
        )

    time_limit_start = 600
    time_limit_repeat = 10

    # first remove any conflicts before ten minutes
    df = df[df['time'] > time_limit_start]

    # now remove any conflicts that don't include mission ids
    df = df[df['ACID1'].isin(mission_acids) | df['ACID2'].isin(mission_acids)]


    # now remove repeat conflicts (10 second)

    # combine the IDS to be able to remove repeat conflicts
    df['ACIDs'] = df.apply(lambda row: tuple(sorted([row['ACID1'], row['ACID2']])), axis=1)
    
    # sort by ACIDS and then time
    df = df.sort_values(by=['ACIDs', 'time'])

    # Iterate over each group of 'combined IDs'
    indices_to_remove = []
    for _, group_df in df.groupby('ACIDs'):

        # skip single conflicts
        if len(group_df) == 1:
            continue

        # calculate forward difference of repeat conflicts
        group_df['time_diff'] = group_df['time'].diff()
        group_df['time_diff'].fillna(15, inplace=True)

        # save indices values
        indices_to_remove += group_df.index[group_df['time_diff'] <= time_limit_repeat].tolist()

    # Drop rows with indices in the set from the DataFrame
    df = df.drop(index=indices_to_remove)

    # get number of conflicts
    n_confs_total = len(df)

    return n_confs_total

def loslog(concept:str, density:str, cluster:str, replanlimit:str, replanratio:str, graphweights:str, seed:str, CRcase:str, logtype:str, mission_acids) -> int:


    scenario_path = scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype)

    df = pd.read_csv(
        scenario_path,
        skiprows=9,
        names=cfg.safety_dict[logtype]['columns']
        )

    time_limit_start = 600
    time_limit_repeat = 10

    # first remove any conflicts before ten minutes
    df = df[df['start_time'] > time_limit_start]

    # remove non mission acids
    df = df[df['ACID1'].isin(mission_acids) | df['ACID2'].isin(mission_acids)]


    # now remove repeat conflicts (10 second)

    # combine the IDS to be able to remove repeat conflicts
    df['ACIDs'] = df.apply(lambda row: tuple(sorted([row['ACID1'], row['ACID2']])), axis=1)
    
    # sort by ACIDS and then time
    df = df.sort_values(by=['ACIDs', 'start_time'])

    # Iterate over each group of 'combined IDs'
    indices_to_remove = []
    for _, group_df in df.groupby('ACIDs'):

        # skip single conflicts
        if len(group_df) == 1:
            continue

        # calculate forward difference of repeat conflicts
        group_df['time_diff'] = group_df['start_time'].diff()
        group_df['time_diff'].fillna(15, inplace=True)

        # save indices values
        indices_to_remove += group_df.index[group_df['time_diff'] <= time_limit_repeat].tolist()

    # Drop rows with indices in the set from the DataFrame
    df = df.drop(index=indices_to_remove)

    # get number of conflicts
    n_intrusions_total = len(df)

    return n_intrusions_total

def flowlog(concept:str, density:str, cluster:str, replanlimit:str, replanratio:str, graphweights:str, seed:str, CRcase:str, logtype:str) -> int:
    """Function returns total flight time and distance.

    Args:
        concept (str): name of concept
        density (str): traffic level
        cluster (str): cluster distnace
        seed (str): random seed number
        logtype (str): log type

    Returns:
        flight_time (float): Returns total flight time in scenario.
        flight_distance (float): Returns total flight time in distance.

    """

    if concept == 'baseline':
        return 0,0,0,0,0

    scenario_path = scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype)

    df = pd.read_csv(
        scenario_path,
        skiprows=9,
        names=cfg.flow_dict['FLOWLOG']['columns']
        )

    # filter time
    df = df[df['time']> 600]

    mean_attempted_replans = df['N_attempted_replans'].mean()
    mean_succesful_replans = df['N_succesful_replans'].mean()
    n_succesful_replans = df['N_succesful_replans'].sum()
    percent_succesful_attempted = df['percent_succesful_attempted'].mean()
    percent_succesful_nair = df['percent_succesful_nair'].mean()
    
    return mean_attempted_replans, mean_succesful_replans, percent_succesful_attempted, percent_succesful_nair, n_succesful_replans

def flstlog(concept:str, density:str, cluster:str, replanlimit:str, replanratio:str, graphweights:str, seed:str, CRcase:str, logtype:str, mission_acids) -> int:
    """Function returns total flight time and distance.

    Args:
        concept (str): name of concept
        density (str): traffic level
        cluster (str): cluster distnace
        seed (str): random seed number
        logtype (str): log type

    Returns:
        flight_time (float): Returns total flight time in scenario.
        flight_distance (float): Returns total flight time in distance.

    """


    scenario_path = scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype)


    df = pd.read_csv(
        scenario_path,
        skiprows=9,
        names=cfg.efficency_dict['FLSTLOG']['columns']
        )

    # remove anything before 10 mins
    df = df[df['spawn_time'] > 600]    

    # remove non mission acids
    df = df[df['ACID'].isin(mission_acids)]

    n_flights = len(df)
    flight_time = df['flight_time'].sum()
    flight_distance = df['distance_2D'].sum()

    return n_flights, flight_time, flight_distance

def clusterlog(concept:str, density:str, cluster:str, replanlimit:str, replanratio:str, graphweights:str, seed:str, CRcase:str, logtype:str) ->  gpd.GeoDataFrame:

    total_airspace_length = 604875
    if concept == 'baseline':
        return 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

    # get scenario
    scenario_path = scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype)

    # initialize the date
    date = datetime.strptime(f'2024-01-01',"%Y-%m-%d")
    
    # intialize data
    data = {
        'time': [],
        'observationcount': [],
        'geometry': [],
        'clusterlength': [],
        'clusterdensity': [],
        'clustercategory': []
    }

    # first delete the first nine lines and strip new line text
    with open(scenario_path) as f:
        lines = f.readlines()
    lines = lines[9:]

    lines = [line.strip('\n') for line in lines]

    # for loop through 5 lines at a time
    for i in range(0, len(lines), 5):
        chunk = lines[i:i+5]
        sec_sim =  float(chunk[0].split(',')[0])

        # skip first ten minutes
        if sec_sim < 600:
            continue

        time_stamp = str(date + timedelta(seconds=sec_sim))

        # get the counts
        counts = chunk[0].split(',')[1:]

        # get polygons
        polygons = chunk[1].split('0,')[-1]

        # now replace ),P' with ')-P' and split with -
        polygons = polygons.replace('),P', ')-P')
        polygons = polygons.split('-')
        # convert to shapely geometry
        polygons = [loads(polygon) for polygon in polygons]

        # get rest of data
        cluster_lengths = chunk[2].split(',')[1:]
        linear_densities = chunk[3].split(',')[1:]
        density_categories = chunk[4].split(',')[1:]

        dataframe_entries = zip(counts, polygons, cluster_lengths, linear_densities, density_categories)
        

        for entry in dataframe_entries:
            data['time'].append(time_stamp)
            data['observationcount'].append(float(entry[0]))
            data['geometry'].append(entry[1])
            data['clusterlength'].append(float(entry[2]))
            data['clusterdensity'].append(float(entry[3]))
            data['clustercategory'].append(entry[4])

    # create gdf        
    gdf = gpd.GeoDataFrame(data, crs='epsg:28992')

    # ensure it is datetime
    gdf['time'] = pd.to_datetime(gdf['time'])

    # calculate hausdorff distance for medium clusters
    relevant_clusters_medium = gdf[gdf['clustercategory'] == 'medium']

    if len(relevant_clusters_medium):
        hausdorff_distances_medium = []
        overlap_percentages_medium = []

        polygon_vectors = []
        multipolygons = []

        for _, group in relevant_clusters_medium.groupby('time'):
            xy_vector = np.column_stack((group.geometry.centroid.x, group.geometry.centroid.y))
            polygon_vectors.append(xy_vector)
            multipolygons.append(group.unary_union)


        # Iterate through the list of 2D vectors
        for i in range(1, len(polygon_vectors)):
            # Calculate Hausdorff distance between consecutive pairs of vectors
            hausdorff_distances = directed_hausdorff(polygon_vectors[i], polygon_vectors[i-1])[0]
            hausdorff_distances_medium.append(hausdorff_distances)
            current_polygon = multipolygons[i]
            prev_polygon = multipolygons[i-1]

            intersection_area = current_polygon.intersection(prev_polygon).area
            total_area = current_polygon.area + prev_polygon.area - intersection_area
            overlap_percent = (intersection_area / total_area) * 100
            overlap_percentages_medium.append(overlap_percent)

        overlap_percentages_medium = np.array(overlap_percentages_medium).mean()
        hausdorff_distances_medium = np.array([hausdorff_distances_medium]).mean()
    else:
        hausdorff_distances_medium = 0
        overlap_percentages_medium = 0


    # calculate hausdorff distance for medium clusters
    relevant_clusters_high = gdf[gdf['clustercategory'] == 'high']

    if len(relevant_clusters_high):
        hausdorff_distances_high = []
        overlap_percentages_high = []

        polygon_vectors = []
        multipolygons = []

        for _, group in relevant_clusters_high.groupby('time'):
            xy_vector = np.column_stack((group.geometry.centroid.x, group.geometry.centroid.y))
            polygon_vectors.append(xy_vector)
            multipolygons.append(group.unary_union)


        # Iterate through the list of 2D vectors
        for i in range(1, len(polygon_vectors)):
            # Calculate Hausdorff distance between consecutive pairs of vectors
            hausdorff_distances = directed_hausdorff(polygon_vectors[i], polygon_vectors[i-1])[0]
            hausdorff_distances_high.append(hausdorff_distances)

            current_polygon = multipolygons[i]
            prev_polygon = multipolygons[i-1]

            # get overlap area
            intersection_area = current_polygon.intersection(prev_polygon).area
            total_area = current_polygon.area + prev_polygon.area - intersection_area
            overlap_percent = (intersection_area / total_area) * 100
            overlap_percentages_high.append(overlap_percent)

        overlap_percentages_high = np.array(overlap_percentages_high).mean()
        hausdorff_distances_high = np.array([hausdorff_distances_high]).mean()
    else:
        hausdorff_distances_high = 0
        overlap_percentages_high = 0

    hausdorff_distances_all = (hausdorff_distances_high + hausdorff_distances_medium)/2
    overlap_percentages_all = (overlap_percentages_high + overlap_percentages_medium)/2

    # get some overview data
    category_counts = gdf['clustercategory'].value_counts()

    # get percent of clusters to verify
    percent_low_clusters = (category_counts.get('low', 0) / len(gdf))*100
    percent_medium_clusters = (category_counts.get('medium', 0) / len(gdf))*100
    percent_high_clusters = (category_counts.get('high', 0) / len(gdf))*100

    # get average length labeled low
    sum_lengths = gdf.groupby(['time','clustercategory'])['clusterlength'].sum()
    percent_length = sum_lengths * (100/total_airspace_length)
    average_percent_length = percent_length.groupby('clustercategory').mean()

    percent_medium_airspace = average_percent_length.get('medium', 0)
    percent_high_airspace = average_percent_length.get('high', 0)
    percent_clustered_airspace = average_percent_length.get('low', 0) + percent_medium_airspace + percent_high_airspace

    return percent_medium_clusters, percent_high_clusters, percent_medium_airspace, \
            percent_high_airspace, percent_clustered_airspace, \
            hausdorff_distances_medium, hausdorff_distances_high, hausdorff_distances_all, \
            overlap_percentages_medium, overlap_percentages_high, overlap_percentages_all

def replanlog(concept:str, density:str, cluster:str, replanlimit:str, replanratio:str, graphweights:str, seed:str, CRcase:str, logtype:str) ->  gpd.GeoDataFrame:


    # get scenario
    scenario_path = scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype)

    # intialize data
    data = {
        'acid': [],
        'number_replans': [],
        'no_replans': [],
        'executed_replans': [],

    }

    # first delete the first nine lines and strip new line text
    with open(scenario_path) as f:
        lines = f.readlines()
    lines = lines[9:]

    lines = [line.strip('\n') for line in lines]

    weird_acids = []
    cluster_zone_acids = []
    mission_acids = []
    # for loop through 5 lines at a time
    for i in range(0, len(lines), 6):
        chunk = lines[i:i+6]
        sec_sim =  float(chunk[0].split(',')[0])

        # skip first ten minutes
        if sec_sim < 600:
            continue
        
        # get acid
        acid = chunk[0].split(',')[1]

        # get type of mission
        misson_type = chunk[0].split(',')[-1]


        if misson_type == 'clusterzone':
            cluster_zone_acids.append(acid)
            continue
        else:
            mission_acids.append(acid)

        # get original plan
        original_plan = chunk[1].split(',')[-1]

        # get travelled plan
        travelled_plan = chunk[2].split(',')[-1]

        # check if they are the same
        is_same_plan = chunk[3].split(',')[-1]
        is_same_plan = True if is_same_plan == 'True' else False

        # get number of replans
        number_replans = int(chunk[4].split(',')[-1])
        
        # process the replans
        replans = chunk[5].split(',')[2:]

        if len(replans) > 1:

            # format them nicely
            formatted_replans = []
            for x_plan in replans:
                formatted_replans.append([int(x) for x in x_plan.split('-')])
            replans = formatted_replans
          
        # check the how many replans were followed
        if not is_same_plan and len(replans) == 1:
        
            # if not the same route and length of replans is 1 then the replan was respected
            executed_replans = 1

        elif not is_same_plan and len(replans) > 1:
 
            # not the same plan and multiple replans

            original_plan = [int(x) for x in original_plan.split('-')]
            travelled_plan = [int(x) for x in travelled_plan.split('-')]

            # this is the number of replans that were followed
            executed_replans = 0

            for idx, replan in enumerate(replans):

                # cut the travelled plan from replan node
                travelled_subset = travelled_plan[travelled_plan.index(replan[0]):]

                # first check is to see if the travelled subset is equal to the replan
                # if yes then the future plans were not performed
                if travelled_subset == replan:
                    executed_replans += 1
                    break

                else:
                    current_start_node = replan[0]
                    # first get the previous plan              
                    if idx == 0:
                        old_plan = original_plan
                    
                    # cut the old plan to start at current node
                    old_plan = old_plan[old_plan.index(current_start_node):]

                    # get part of replan that is not in old plan
                    new_replan_part = [replan[i] for i, item in enumerate(replan) if item not in old_plan]

                    # get next replan and its first node
                    try:
                        next_replan = replans[idx + 1]
                    except:
                        weird_acids.append(acid)
                        break
                    first_node_next_replan = next_replan[0]

                    # with this we can rebuild the future travelled route
                    future_travelled_route = travelled_plan[travelled_plan.index(current_start_node):travelled_plan.index(first_node_next_replan)]

                    # now check if any nodes in the new replan part are in the future travelled route
                    check_future_nodes = any(item in future_travelled_route for item in new_replan_part)

                    # if true this was a succesful replan
                    if check_future_nodes:
                        executed_replans += 1

            
                    # at end of replan for loop
                    old_plan = deepcopy(replan)
        else:
            executed_replans = 0
        

        # create dictionary
        data['acid'].append(acid)
        data['number_replans'].append(number_replans)
        data['no_replans'].append(is_same_plan)
        data['executed_replans'].append(executed_replans)

    if concept == 'baseline':
        return 0, 0, 0, 0, 0, 0, mission_acids
    
    # create gdf        
    df = pd.DataFrame(data)

    number_of_weird_situations = 0
    if len(weird_acids) > 0:
        # print(scenario_path)
        # print(weird_acids)
        number_of_weird_situations = round(100*len(weird_acids)/len(df),2)
        if number_of_weird_situations > 1:
            print(scenario_path)
        # print(round(100*len(weird_acids)/len(df),2))
        # print('------------------------')

    # get some overview data
    no_replan = df['no_replans'].value_counts()

    # get some stats
    ac_number_of_replanning_aircraft = no_replan.get(False, 0)
    ac_number_of_not_replanning_aircraft = no_replan.get(True, 0)

    # number of aircraft that attempted replans
    ac_number_of_attempted_replans  = (df['number_replans'] > 0).sum()

    # get followed replans
    executed_replans_per_flight = df['executed_replans'].sum() / len(df)
    attempted_replans_per_flight = df['number_replans'].sum() / len(df)
    # print(cluster,density)
    # print(round(100*(executed_replans_per_flight/attempted_replans_per_flight),2))
    
    return ac_number_of_replanning_aircraft, ac_number_of_not_replanning_aircraft, ac_number_of_attempted_replans, \
            executed_replans_per_flight, attempted_replans_per_flight, number_of_weird_situations, mission_acids

def reglog(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype, mission_acids) -> gpd.GeoDataFrame:
    """Function returns total flight time and distance.

    Args:
        concept (str): name of concept
        density (str): traffic level
        cluster (str): cluster distnace
        seed (str): random seed number
        logtype (str): log type

    Returns:
        flight_time (float): Returns total flight time in scenario.
        flight_distance (float): Returns total flight time in distance.

    """
    scenario_path = scenario_choice(concept, density, cluster, replanlimit, replanratio, graphweights, seed, CRcase, logtype)


    # read the files and skip the first 9 rows
    header_columns = ['percent_missionACIDs','percent_clusterACIDs','ALTs','LATs','LONs']
    date = datetime.strptime(f'2024-01-01',"%Y-%m-%d")
    data = dict()
    time_stamp_entries = dict()

    with open(scenario_path) as f:
        for line_num, line in enumerate(f):

            if line_num < 9:
                continue
            header_id = (line_num - 9) % 4


            # get the time
            sec_sim = float(line.split(',')[0])

            if sec_sim <= 600:
                    continue
        
            time_stamp = str(date + timedelta(seconds=sec_sim))
            if header_id == 0:
                # make a dictionary wth values for each header_column
                data[time_stamp] = {header_col:[] for header_col in header_columns}

                # get acids that are in mission type
                acids = set(line.strip('\n').split(',')[1:])
                n_flights = len(acids)
                n_mission_acids = len(list(acids & set(mission_acids)))

                # get the cluster acids
                n_cluster_acids = n_flights - n_mission_acids
                

                data[time_stamp]['percent_missionACIDs'] = n_mission_acids / n_flights * 100
                data[time_stamp]['percent_clusterACIDs'] = n_cluster_acids

            else:
                data[time_stamp][header_columns[header_id+1]] = [float(i) for i in line.strip('\n').split(',')[1:]]

            # create a check to see if the data is complete.
            time_stamp_entries[time_stamp] = time_stamp_entries.get(time_stamp, 0) + 1

    # get a dataframe of the entries
    df_time_entries = pd.DataFrame(time_stamp_entries, index=['count']).T

    # convert to pandas dataframe where the index is the data keys and the columns are the header columns
    df = pd.DataFrame(data).T

    # remove entries from df that have a value less than 4 in the df_time_entries
    df = df[df_time_entries['count'] >= 4]

    multi_point = np.array([MultiPoint(np.column_stack((x, y))) for x,y in zip(df['LONs'], df['LATs'])], dtype=MultiPoint)

    # remove ACID, LAT, LON, ALT columns
    df.drop(columns=['LATs', 'LONs', 'ALTs'], inplace=True)

    # make a geodataframe from the pandas dataframe
    gdf = gpd.GeoDataFrame(df, geometry=multi_point, crs='epsg:4326')

    # set the index as a column
    gdf['date'] = gdf.index
    gdf['date']= pd.to_datetime(gdf['date'])

    # reset index
    gdf.reset_index(inplace=True, drop=True)

    # convert to epsg 28992
    gdf = gdf.to_crs(epsg=28992)

    # explode df
    # single_parts_gdf = gdf.explode(ignore_index=True)

    # get average of mission aircraft
    percent_mission_average = gdf['percent_missionACIDs'].mean()

    print(percent_mission_average)

    return percent_mission_average

if __name__ == '__main__':
    main()
